import { createAction, props, UPDATE, INIT, META_REDUCERS } from '@ngrx/store';
import * as i0 from '@angular/core';
import { InjectionToken, Injectable, Inject, APP_INITIALIZER, NgModule, Optional, SkipSelf } from '@angular/core';
import * as i1 from '@ngrx/effects';
import { createEffect, ofType, EffectsModule } from '@ngrx/effects';
import { __awaiter } from 'tslib';
import deepmerge from 'deepmerge';
import { set, createStore, get, update } from 'idb-keyval';
import { take, mergeMap, tap, map, catchError, switchMap } from 'rxjs/operators';
import { from, of, ReplaySubject, timer, EMPTY } from 'rxjs';

const rehydrateInitAction = createAction('NgrxStoreIdb/Init');
const rehydrateAction = createAction('NgrxStoreIdb/Rehydrate', props());
const rehydrateErrorAction = createAction('NgrxStoreIdb/RehydrateError');

var ngrxStoreIdb_actions = /*#__PURE__*/Object.freeze({
    __proto__: null,
    rehydrateInitAction: rehydrateInitAction,
    rehydrateAction: rehydrateAction,
    rehydrateErrorAction: rehydrateErrorAction
});

/**
 * Injection token for injection options
 */
const OPTIONS = new InjectionToken('NgrxStoreIdb options');
/**
 * Injection token for injecting IDB store
 */
const IDB_STORE = new InjectionToken('IDB Store');
/**
 * Name of the key in IndexedDB database store under which the state will be saved
 */
const SAVED_STATE_KEY = 'State';

/**
 * Default marshaller saves the whole store state
 */
const defaultMarshaller = (state) => state;
// Default merge strategy is a full deep merge.
const defaultUnmarshaller = (state, rehydratedState) => {
    const overwriteMerge = (destinationArray, sourceArray, options) => sourceArray;
    const options = {
        arrayMerge: overwriteMerge,
    };
    return deepmerge(state, rehydratedState, options);
};
const DEFAULT_OPTS = {
    rehydrate: true,
    saveOnChange: true,
    syncCondition: null,
    keys: null,
    unmarshaller: defaultUnmarshaller,
    marshaller: defaultMarshaller,
    debugInfo: true,
    idb: {
        dbName: 'NgrxStoreIdb',
        storeName: 'Store',
    },
    concurrency: {
        allowed: false,
        refreshRate: 5000,
        trackKey: 'ConcurrencyTimestamp',
        acquireLockOnStartup: true,
        failInitialisationIfNoLock: false,
    },
};
// Recursively traverse all properties of the existing slice as defined by the `filter` argument,
// and output the new object with extraneous properties removed.
const createStateSlice = (existingSlice, keys) => {
    return keys.reduce((memo, attr) => {
        if (typeof attr === 'string' || typeof attr === 'number') {
            const value = existingSlice === null || existingSlice === void 0 ? void 0 : existingSlice[attr];
            if (value !== undefined) {
                memo[attr] = value;
            }
        }
        else {
            for (const key in attr) {
                if (Object.prototype.hasOwnProperty.call(attr, key)) {
                    const element = attr[key];
                    memo[key] = createStateSlice(existingSlice[key], element);
                }
            }
        }
        return memo;
    }, {});
};
/**
 * Marshalls actual state using keys
 */
const keysMarshaller = (state, keys) => {
    if (!state) {
        return state;
    }
    const res = {};
    keys.forEach(key => {
        let name = key;
        // If key was a string then stateSlice has value and we are done
        let stateSlice = state[name];
        // If key was a nested structure then we need to dig deeper
        if (typeof key === 'object') {
            // keys configuration always has only 1 attribute and the rest is nested
            name = Object.keys(key)[0];
            stateSlice = state[name];
            stateSlice = createStateSlice(stateSlice, key[name]);
        }
        if (stateSlice !== undefined) {
            res[name] = stateSlice;
        }
    });
    return res;
};
let lastSavedState = null;
/**
 * Compare two objects if they are equal.
 * They are equal if both have keys with the same name
 * and the same value.
 */
const statesAreEqual = (prev, next) => {
    if (prev === next) {
        return true;
    }
    if (!prev || !next) {
        return false;
    }
    if (typeof prev !== 'object' || typeof next !== 'object') {
        return false;
    }
    const prevSlices = Object.keys(prev);
    const nextSlices = Object.keys(next);
    if (prevSlices.length !== nextSlices.length) {
        return false;
    }
    for (const slice of prevSlices) {
        if (!statesAreEqual(prev[slice], next[slice])) {
            return false;
        }
    }
    return true;
};
/**
 * Method used to save actual state into IndexedDB
 */
const syncStateUpdate = (state, action, opts, idbStore, service) => {
    if (!service.canConcurrentlySync()) {
        if (opts.debugInfo) {
            if (opts.concurrency.acquireLockOnStartup) {
                console.debug('NgrxStoreIdb: State will not be persisted. Application runs also in other tab/window.');
            }
            else {
                console.debug('NgrxStoreIdb: State will not be persisted. Concurrency lock currently not acquired.');
            }
        }
        return;
    }
    if (opts.syncCondition) {
        try {
            if (opts.syncCondition(state, action) !== true) {
                if (opts.debugInfo) {
                    console.debug('NgrxStoreIdb: State will not be persisted. syncCondition is false');
                }
                return;
            }
        }
        catch (e) {
            // Treat TypeError as do not sync
            if (e instanceof TypeError) {
                if (opts.debugInfo) {
                    console.debug('NgrxStoreIdb: State will not be persisted. syncCondition has error', e);
                }
                return;
            }
            console.error('NgrxStoreIdb: syncCondition raised error', e);
            throw e;
        }
    }
    let marshalledState = {};
    if (opts.keys) {
        marshalledState = keysMarshaller(state, opts.keys);
    }
    else {
        marshalledState = opts.marshaller(state);
    }
    if (opts.saveOnChange && statesAreEqual(lastSavedState, marshalledState)) {
        if (opts.debugInfo) {
            console.debug('NgrxStoreIdb: No change in state. Will skip saving to IndexedDB.');
        }
        return;
    }
    set(SAVED_STATE_KEY, marshalledState, idbStore)
        .then(() => {
        lastSavedState = marshalledState;
        service.broadcastSyncEvent(action);
        if (opts.debugInfo) {
            console.debug('NgrxStoreIdb: Store state persisted to IndexedDB', marshalledState, action);
        }
    })
        .catch(err => {
        if (opts.debugInfo) {
            console.error('NgrxStoreIdb: Error storing state to IndexedDB', err, action);
        }
    });
};
/**
 * This is the main factory that creates our metareducer.
 */
const metaReducerFactoryWithOptions = (options, idbStore, service) => {
    let rehydratedState = null;
    return (reducer) => (state, action) => {
        let nextState;
        if (options.debugInfo) {
            console.group('NgrxStoreIdb: metareducer', state, action);
        }
        // If we are processing rehydrateAction then save rehydrated state (for later use).
        // There is no other reducer for this action.
        if (action.type === rehydrateAction.type) {
            const payload = action;
            rehydratedState = payload.rehydratedState || {};
            if (!payload.rehydratedState) {
                if (options.debugInfo) {
                    console.debug('NgrxStoreIdb: Rehydrated state is empty - nothing to rehydrate.');
                    console.groupEnd();
                }
                return state;
            }
        }
        // If action is rehydrateAction (i.e. initial rehydratation)
        // then merge the store state with the rehydrated state
        if (action.type === rehydrateAction.type) {
            nextState = options.unmarshaller(state, rehydratedState);
            service.broadcastRehydrateEvent(nextState);
            if (options.debugInfo) {
                console.debug('NgrxStoreIdb: After rehydrating current state', nextState);
            }
        }
        else {
            // Run normal reducer for this action
            nextState = reducer(state, action);
        }
        // If action is UPDATE then rehydrate feature slices just created (when lazy module store loads)
        if (action.type === UPDATE && action.features && rehydratedState) {
            const rehydratedStateCopy = {};
            for (const feature of action.features) {
                if (rehydratedState[feature]) {
                    rehydratedStateCopy[feature] = rehydratedState[feature];
                }
            }
            nextState = options.unmarshaller(nextState, rehydratedStateCopy);
            service.broadcastRehydrateEvent(nextState);
            if (options.debugInfo) {
                console.debug('NgrxStoreIdb: After rehydrating current state', nextState);
            }
        }
        if (action.type !== INIT &&
            action.type !== UPDATE &&
            action.type !== rehydrateInitAction.type &&
            action.type !== rehydrateAction.type &&
            action.type !== rehydrateErrorAction.type &&
            // If rehydrating is requested then don't sync until the saved state was loaded first
            (rehydratedState || !options.rehydrate)) {
            if (options.debugInfo) {
                console.debug('NgrxStoreIdb: Try to persist state into IndexedDB', nextState, action);
            }
            syncStateUpdate(nextState, action, options, idbStore, service);
        }
        if (options.debugInfo) {
            console.groupEnd();
        }
        return nextState;
    };
};
const optionsFactory = (options) => {
    if (options.keys && (options.unmarshaller || options.marshaller)) {
        throw new Error('NgrxStoreIdb: define keys or unmarshaller+marshaller but not both!');
    }
    if (!!options.marshaller !== !!options.unmarshaller) {
        throw new Error('NgrxStoreIdb: define unmarshaller and marshaller!');
    }
    const opts = deepmerge(DEFAULT_OPTS, options);
    if (opts.debugInfo) {
        console.info('NgrxStoreIdbModule: Using the following options', Object.assign(Object.assign({}, opts), { marshaller: opts.marshaller === defaultMarshaller ? 'default marshaller' : 'custom marshaller', unmarshaller: opts.unmarshaller === defaultUnmarshaller ? 'default unmarshaller' : 'custom unmarshaller' }));
    }
    return opts;
};
const idbStoreFactory = (opts) => {
    return createStore(opts.idb.dbName, opts.idb.storeName);
};
const ngrxStoreIdbServiceInitializer = (opts, service) => {
    return () => __awaiter(void 0, void 0, void 0, function* () {
        if (opts.rehydrate) {
            yield service.onRehydrate().pipe(take(1)).toPromise();
        }
        if (opts.concurrency.acquireLockOnStartup) {
            const hasLock = yield service.onLockAcquired().toPromise();
            if (hasLock || !opts.concurrency.failInitialisationIfNoLock) {
                return true;
            }
            else {
                throw new Error('Can not acquire master lock. Another tab/window is open?');
            }
        }
        else {
            return true;
        }
    });
};

class RehydrateEffects {
    constructor(actions$, options, idbStore) {
        this.actions$ = actions$;
        this.options = options;
        this.idbStore = idbStore;
        /**
         * This effect is triggered after root store initialisation.
         * This is the very first thing done by NgRx and the best place
         * to load saved data.
         */
        this.rehydrateOnInit$ = createEffect(() => this.actions$.pipe(ofType(rehydrateInitAction), mergeMap(action => {
            if (this.options.debugInfo) {
                console.debug('NgrxStoreIdb: Load state from IndexedDB');
            }
            return from(get(SAVED_STATE_KEY, this.idbStore)).pipe(tap(value => this.rehydratedState = value), tap(value => {
                if (this.options.debugInfo) {
                    console.debug('NgrxStoreIdb: Loaded state from IndexedDB:', value);
                }
            }), map(value => rehydrateAction({ rehydratedState: value })), catchError(err => {
                console.error('NgrxStoreIdb: Error reading state from IndexedDB', err);
                return of(rehydrateErrorAction());
            }));
        })));
    }
    ngrxOnInitEffects() {
        if (this.options.rehydrate) {
            return rehydrateInitAction();
        }
    }
}
RehydrateEffects.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: RehydrateEffects, deps: [{ token: i1.Actions }, { token: OPTIONS }, { token: IDB_STORE }], target: i0.ɵɵFactoryTarget.Injectable });
RehydrateEffects.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: RehydrateEffects, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: RehydrateEffects, decorators: [{
            type: Injectable,
            args: [{
                    providedIn: 'root',
                }]
        }], ctorParameters: function () {
        return [{ type: i1.Actions }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [OPTIONS]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [IDB_STORE]
                    }] }];
    } });

/**
 * This service emits events each time NgrxStoreIdb metareducer successfuly
 * syncs data to IndexedDB. The data emited from onSync$ observable is
 * the action that triggered the synchronisation event.
 */
class NgrxStoreIdbService {
    constructor(opts, idbStore) {
        this.opts = opts;
        this.idbStore = idbStore;
        this.syncSubject = new ReplaySubject(1);
        this.onSync$ = this.syncSubject.asObservable();
        this.rehydrateSubject = new ReplaySubject(1);
        this.onRehydrate$ = this.rehydrateSubject.asObservable();
        this.lockAcquiredSubject = new ReplaySubject(1);
        this.onLockAcquired$ = this.lockAcquiredSubject.asObservable();
        this.iAmMasterOfStore = false;
        this.uniqueKey = this.uuidv4();
        if (opts.concurrency.acquireLockOnStartup) {
            this.acquireLock(null).subscribe(acquired => {
                if (this.opts.debugInfo) {
                    if (acquired) {
                        console.debug('NgrxStoreIdb: Succesfully acquired lock on startup');
                    }
                    else {
                        console.debug('NgrxStoreIdb: Could not acquired lock on startup');
                    }
                }
            });
        }
        else {
            this.lockAcquiredSubject.next(false);
        }
    }
    acquireLock(storeToRehydrate) {
        if (this.iAmMasterOfStore)
            return of(true);
        return from(update(this.opts.concurrency.trackKey, data => {
            if (!data || data.timestamp < (Date.now() - this.opts.concurrency.refreshRate * 1.1)) {
                return {
                    uniqueKey: this.uniqueKey,
                    timestamp: Date.now(),
                };
            }
            else {
                return data;
            }
        }, this.idbStore)).pipe(switchMap(() => from(get(this.opts.concurrency.trackKey, this.idbStore)).pipe(
        // Have a look if there is already some other instance running
        map(inData => !inData || inData.uniqueKey === this.uniqueKey), switchMap(lockAcquired => {
            if (lockAcquired) {
                this.iAmMasterOfStore = true;
                this.lockAcquiredSubject.next(true);
                // If manual lock managenent is disabled, lock status will not change anymore
                if (this.opts.concurrency.acquireLockOnStartup) {
                    this.lockAcquiredSubject.complete();
                }
                // No instance or it was not updated for a long time.
                // Start a timer and keep updating the timestamp
                this.timerSubscription = timer(0, this.opts.concurrency.refreshRate).pipe(map(() => ({
                    uniqueKey: this.uniqueKey,
                    timestamp: Date.now(),
                })), switchMap(outData => {
                    if (this.iAmMasterOfStore) {
                        return from(set(this.opts.concurrency.trackKey, outData, this.idbStore)).pipe(map(() => outData));
                    }
                    else {
                        return EMPTY;
                    }
                })).subscribe(outData => {
                    if (this.opts.debugInfo) {
                        console.debug(`NgrxStoreIdb: Updating concurrency timestamp '${this.opts.concurrency.trackKey}'`, outData);
                    }
                });
                if (storeToRehydrate) {
                    return from(get(SAVED_STATE_KEY, this.idbStore)).pipe(tap(value => {
                        storeToRehydrate.dispatch(rehydrateAction({ rehydratedState: value }));
                        if (this.opts.debugInfo) {
                            console.debug('NgrxStoreIdb: Loaded state from IndexedDB:', value);
                        }
                    }), map(() => true), catchError(err => {
                        console.error('NgrxStoreIdb: Error reading state from IndexedDB', err);
                        storeToRehydrate.dispatch(rehydrateErrorAction());
                        return of(false);
                    }));
                }
                else {
                    return of(true);
                }
            }
            else {
                // Otherwise do nothing - some other instance is syncing/master of the IDB store
                this.iAmMasterOfStore = false;
                this.lockAcquiredSubject.next(false);
                // If manual lock managenent is disabled, lock status will not change anymore
                if (this.opts.concurrency.acquireLockOnStartup) {
                    this.lockAcquiredSubject.complete();
                }
                return of(false);
            }
        }))));
    }
    releaseLock() {
        if (!this.iAmMasterOfStore)
            return EMPTY;
        this.timerSubscription.unsubscribe();
        this.timerSubscription = null;
        this.iAmMasterOfStore = false;
        this.lockAcquiredSubject.next(false);
        return from(update(this.opts.concurrency.trackKey, data => {
            if (!data || data.uniqueKey === this.uniqueKey) {
                return null;
            }
            else {
                return data;
            }
        }, this.idbStore));
    }
    onLockAcquired() {
        return this.onLockAcquired$;
    }
    canConcurrentlySync() {
        return this.opts.concurrency.allowed || this.iAmMasterOfStore;
    }
    uuidv4() {
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, (c) => {
            const r = Math.random() * 16 | 0, v = c === 'x' ? r : (r & 0x3 | 0x8);
            return v.toString(16);
        });
    }
    broadcastSyncEvent(action) {
        this.syncSubject.next(action);
    }
    broadcastRehydrateEvent(state) {
        this.rehydrateSubject.next(state);
    }
    onSync() {
        return this.onSync$;
    }
    onRehydrate() {
        return this.onRehydrate$;
    }
}
NgrxStoreIdbService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbService, deps: [{ token: OPTIONS }, { token: IDB_STORE }], target: i0.ɵɵFactoryTarget.Injectable });
NgrxStoreIdbService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbService, decorators: [{
            type: Injectable,
            args: [{
                    providedIn: 'root',
                }]
        }], ctorParameters: function () {
        return [{ type: undefined, decorators: [{
                        type: Inject,
                        args: [OPTIONS]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [IDB_STORE]
                    }] }];
    } });

/**
 * Import this module in your AppModule using forRoot() method to
 * enable synchornisation of redux store with IndexedDB.
 */
class NgrxStoreIdbModule {
    constructor(parentModule) {
        if (parentModule) {
            throw new Error('NgrxStoreIdbModule is already loaded. Import it in the AppModule only');
        }
    }
    static forRoot(options = {}) {
        return {
            ngModule: NgrxStoreIdbModule,
            providers: [
                // Used to pass options into RehydrateEffects
                {
                    provide: OPTIONS,
                    useValue: optionsFactory(options),
                },
                // Used to pass idb store into RehydrateEffects
                {
                    provide: IDB_STORE,
                    deps: [OPTIONS],
                    useFactory: idbStoreFactory,
                },
                {
                    provide: APP_INITIALIZER,
                    deps: [OPTIONS, NgrxStoreIdbService],
                    useFactory: ngrxStoreIdbServiceInitializer,
                    multi: true,
                },
                // This installs NgrxStateIdb metareducer into use
                {
                    provide: META_REDUCERS,
                    deps: [OPTIONS, IDB_STORE, NgrxStoreIdbService],
                    useFactory: metaReducerFactoryWithOptions,
                    multi: true,
                },
            ],
        };
    }
}
NgrxStoreIdbModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbModule, deps: [{ token: NgrxStoreIdbModule, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.NgModule });
NgrxStoreIdbModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbModule, imports: [i1.EffectsFeatureModule] });
NgrxStoreIdbModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbModule, imports: [[
            EffectsModule.forFeature([RehydrateEffects]),
        ]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.3.11", ngImport: i0, type: NgrxStoreIdbModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [],
                    imports: [
                        EffectsModule.forFeature([RehydrateEffects]),
                    ],
                    exports: [],
                }]
        }], ctorParameters: function () {
        return [{ type: NgrxStoreIdbModule, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }] }];
    } });

/*
 * Public API Surface of ngrx-store-idb
 */

/**
 * Generated bundle index. Do not edit.
 */

export { NgrxStoreIdbModule, NgrxStoreIdbService, ngrxStoreIdb_actions as RehydrateActions };
//# sourceMappingURL=ngrx-store-idb-forked.mjs.map
